/*
 * leds
 * Linux Editor for D 
 * S as in simple(superfluous) or supperb
 * 
 * by Antonio Monteiro
 */

module leds.FSView;

private import leds.Leds;
private import leds.Directory;
private import leds.Workspace;
private import leds.FileView;
private import leds.Docker;
private import leds.Project;
//private import leds.POpen;
private import leds.ExecConsole;

private import dui.All;
private import std.string;
private import std.path;
private import std.file;

private alias std.path filePath;

private class RowType
{
	static const char[] EMPTY = "E";
	static const char[] DIRECTORY = "D";
	static const char[] FILE = "F";
}

extern(C)
{
};

/**
 * Displays content of one directory.
 * column 1 - the visible string to the user
 * column 2 - the row type { DIRECTORY or FILE }
 * column 3 - the value { always the full PATH  }
 */
public:
class FSView : VBox , ButtonClickedListener , Dockable
{
	
	Workspace workspace;

	char[] currentDir;

	FSTreeView treeView;
	ScrolledWindow sw;

	// our tree definition
	/** the columns */
	static GType [3] columns = [DUIType.STRING,DUIType.STRING,DUIType.STRING]; 
	
	/** the tree store */
	FSTreeStore fsTreeStore;
	/** the root iter */
	TreeIter iterTop;
	/** the current projects */
	//Project [char[]] projects;

	Menu tabMenu;

	Docker docker;

	/**
	 * Creates a new project view for a specific workspace
	 * @param workspace the workspace
	 */
	this(Workspace workspace)
	{
		super(false,0);
		this.workspace = workspace;
		
		currentDir = std.file.getcwd();
		
		sw = new ScrolledWindow(PolicyType.AUTOMATIC, PolicyType.AUTOMATIC);
		treeView = createFSTree();
		sw.add(treeView);
		packStart(getToolbar(),false,false,0);
		packStart(sw,true,true,0);
		
		loadCurrentDir();
	}

	Widget getWidget()
	{
		return this;
	}
	
	char[] getTabName()
	{
		return "fs";
	}
	
	Menu getTabMenu()
	{
		if ( tabMenu === null )
		{
			tabMenu = new Menu();
			MenuItem item;
			
			//tabMenu.append(new MenuItem(this,"_Close","tabMenu.close"));
			//tabMenu.append(new MenuItem(this,"_Float","tabMenu.float"));
			tabMenu.append(new MenuItem("test fs"));
		}
		//return tabMenu;
		return null;
	}

	void setDocker(Docker docker)
	{
		this.docker = docker;
	}

	/**
	 * Gets this dockable file view
	 * @return null because with don't have a file view
	 */
	FileView getFileView()
	{
		return null;
	}
	
	/**
	 * detrimines if a close button shoud be visible
	 */
	bit allowCloseButton()
	{
		return false;
	}
	
	bit allowClose()
	{
		return true;
	}

	bit clear()
	{
		return false;
	}

	void loadCurrentDir()
	{
		printf("loading current directory %.*s\n",getCurrentDir());

		TreeIter oldIter = iterTop;
		iterTop = fsTreeStore.createIter();

		static int [3] cols = [0,1,2];
		char [][] subVals;
		subVals ~= currentDir;
		subVals ~= RowType.DIRECTORY;
		subVals ~= getCurrentDir();
		fsTreeStore.set(iterTop,cols,subVals);


		loadDir(iterTop, getCurrentDir(), "", true);
		loadFiles(iterTop, getCurrentDir(),  true);

		if ( oldIter !== null )
		{
			fsTreeStore.remove(oldIter);
		}
	}


	bit validDir(char[] dir, char[] subdir, bit showHidden)
	{
		return true;
	}
	
	/**
	 * loads a directory and all it's subdirectories and files not recursively.
	 * 
	 */
	private void loadDir(TreeIter parentIter, char[] dirPath, char[] superDir, bit showHidden)
	{
		static int [3] cols = [0,1,2];
		if ( dirPath.length == 0 ) return;
		try
		{
			Directory dir = new Directory(dirPath);
			foreach(char[]subdir ; dir.getDirNames().sort)
			{
				if ( validDir(superDir,subdir, showHidden) )
				{
					TreeIter iterChild = fsTreeStore.append(parentIter);
					char [][] subVals;
					subVals ~= superDir~subdir;
					subVals ~= RowType.DIRECTORY;
					subVals ~= dirPath;
					fsTreeStore.set(iterChild,cols,subVals);
					TreeIter iterEmpty = fsTreeStore.append(iterChild);
					char [][] emptyVals;
					emptyVals ~= superDir~subdir;
					emptyVals ~= RowType.EMPTY;
					emptyVals ~= ".";
					fsTreeStore.set(iterEmpty,cols,emptyVals);
					//char[] subPath = dirPath~"/"~subdir;
					//loadDir(parentIter, subPath, superDir~subdir~"/", showHidden);
					//loadFiles(iterChild, subPath, showHidden);
				}
			}
		}
		catch ( DirectoryError e)
		{
			printf("FSView load directory error : %.*s\n",e.toString());
		}
	}

	private void loadFiles(TreeIter iterChild, char[] subPath, bit showHidden)
	{
		static int [3] cols = [0,1,2];
		Directory files = new Directory(subPath);
		foreach(char[]subfile ; files.getFileNames().sort)
		{
			if ( subfile.length>0 
				&& subfile[0]!=0 
				//&& ( showHidden || subfile[0]!='.')
				//&& subfile!="." 
				&& subfile!=".." )
			{
				TreeIter iterFile = fsTreeStore.append(iterChild);
				char [][] subVals;
				subVals  ~= subfile;
				subVals  ~= RowType.FILE;
				subVals  ~= subPath~"/"~subfile;
				fsTreeStore.set(iterFile,cols,subVals);
			}
		}
	}

	/**
	 * Creates the simple toolbar for the project view.
	 * @return a Toolbar
	 */
	Toolbar getToolbar()
	{	
		Dispatcher dispatcher = Dispatcher.getDispatcher();
		Toolbar toolbar = new Toolbar();
		//toolbar.setIconSize(IconSize.SMALL_TOOLBAR);
		toolbar.setIconSize(IconSize.MENU);
		toolbar.setStyle(ToolbarStyle.ICONS);
		
		// this exist because of compiler
		// problems on "forward references to classes"
		void addButton(Stock stockID, char[] tooltip, char[] action)
		{
			Button button = (Button)toolbar.insertStock(stockID,tooltip,"",-1);
			dispatcher.addButtonClickedListener(this,button,action);
		}
			
		addButton(Stock.GO_UP,"Up","toolbar.up");
		addButton(Stock.GO_FORWARD,"Enter","toolbar.enter");
		addButton(Stock.OPEN,"Open","toolbar.open");
		
		return toolbar;
	}

	/**
	 * listens to user button clicks
	 * @param button
	 * @param action string
	 */
	void buttonClickedCallback(Button button, char [] action)
	{
		switch(action)
		{
	
			//case "toolbar.new": ; break;
			//case "toolbar.add": addFile(); break;
			//case "toolbar.remove": remove(); break;
				
			default:
				printf("FS view - You press %.*s\n",action);
				break;
		}
	}

	char[] getCurrentDir()
	{
		return currentDir;
	}

	TreeView getTreeView()
	{
		return treeView;
	}

	/**
	 * A TreeStore for a fs
	 */
	class FSTreeStore : TreeStore
	{
		this(GType[] columns)
		{
			super(columns);
		}
	}
	
	FSTreeView createFSTree()
	{
		fsTreeStore = new FSTreeStore(columns);
		//printf("projectTreeStore %X\n",projectTreeStore);
		FSTreeView treeView = new FSTreeView(fsTreeStore,this);
		TreeSelection ts = treeView.getSelection();
		ts.setMode(SelectionMode.MULTIPLE);
		treeView.setHeadersVisible(false);
		TreeViewColumn column = new TreeViewColumn("Elements",new CellRendererText(),"text", 0);
		treeView.appendColumn(column);

		return treeView;
	}

	/**
	 * A TreeView to display a fs structure
	 */
	class FSTreeView : TreeView , MouseButtonListener , MenuItemListener
	{
	
		Menu projectPopupMenu;
		Menu directoryPopupMenu;
		Menu filePopupMenu;
		
		FSView fsView;

		TreeStore treeStore;
		
		this(TreeStore treeStore, FSView fsView)
		{
			super(treeStore);
			this.treeStore = treeStore;
			this.fsView = fsView;
		//printf("treeStore %X\n",this.treeStore);
			Dispatcher d = Dispatcher.getDispatcher();
			directoryPopupMenu = buildDirectoryMenu();
			filePopupMenu = buildFileMenu();
			d.addMouseButtonListener(this,this);
			//d.addMenuItemListener(this,this);
		}

		Menu buildDirectoryMenu()
		{
			Menu menu = new Menu();
			MenuItem item;
			
			menu.append(new MenuItem(this,"_Filter","directoryPopupMenu.filter"));
			menu.append(new MenuItem(this,"_Hide","directoryPopupMenu.hide"));

			return menu;
		}
		
		Menu buildFileMenu()
		{
			Menu menu = new Menu();
			MenuItem item;
			
			menu.append(new MenuItem(this,"_Open","filePopupMenu.open"));
			menu.append(new MenuItem(this,"_Copy","filePopupMenu.copy"));
			menu.append(new MenuItem(this,"_Paste","filePopupMenu.paste"));
			menu.append(new MenuItem(this,"_Cut","filePopupMenu.cut"));
			menu.append(new MenuItem(this,"_Remove","filePopupMenu.remove"));

			return menu;
		}

		bit mouseButtonPressCallback(Widget widget, EventButton eventButton)
		{
			//printf("mouse pressed on project tree view with button %d and time %d\n",eventButton.button(),eventButton.getTime());
			if ( eventButton.button() == 3 )
			{
			}
			else
			{
			}
			return false;   // false : the event is passed to other listeners
		}
		
		bit mouseButtonReleaseCallback(Widget widget, EventButton eventButton)
		{
			//printf("mouse release on project tree view with button %d\n",eventButton.button());
			TreeIter iter = getSelectedIter();
			if ( iter !== null )
			{
				if ( eventButton.button() == 3 )
				{
					switch(iter.getValueString(1))
					{
						case RowType.DIRECTORY:  
							directoryPopupMenu.popup(eventButton.button(),eventButton.getTime());
							break;
							
						case RowType.FILE:  
							filePopupMenu.popup(eventButton.button(),eventButton.getTime());
							break;
						
						default:
							break;
					}
				}
				else if ( eventButton.button() == 2 )
				{
					switch(iter.getValueString(1))
					{
						case RowType.EMPTY:
							break;
						case RowType.DIRECTORY:
						
printf("currentDir before == %.*s\n",fsView.currentDir);
							fsView.currentDir = filePath.join(fsView.currentDir,iter.getValueString(0).dup);
printf("currentDir during == %.*s\n",fsView.currentDir);
							std.file.chdir(fsView.currentDir);
printf("currentDir during == %.*s\n",fsView.currentDir);
							fsView.currentDir = std.file.getcwd();
printf("currentDir after == %.*s\n",fsView.currentDir);
							fsView.loadCurrentDir();
							break;
						
						case RowType.FILE:  
							fsView.workspace.openFile(iter.getValueString(2));
							break;
						
						default:
							break;
					}
				}
				else
				{
					//projectMenu.popup(eventButton.button(),eventButton.getTime());
				}
			}
			return false;   // false : the event is passed to other listeners
		}

		/**
		 * Menu listener callback
		 */
		void activateCallback(MenuItem menuItem, char [] action)
		{
			TreeIter[] iters = getSelectedIters();
			switch(action)
			{
				
				case "filePopupMenu.open":
					foreach ( TreeIter iter ; iters )
					{
						if ( iter !== null && "F"==iter.getValueString(1) )
						{
							fsView.workspace.openFile(iter.getValueString(2));
						}
					}
					break;
					
				default:
					printf("Menu option %.*s\n", action);
					break;
			}
		}
		
		void activateItemCallback(MenuItem menuItem, char [] action)
		{
			switch(action)
			{
				default:
					printf("Menu option %.*s\n", action);
					break;
			}
		}
		
	}
	
}
